--------------------------------------------------
--	RasChess LUA 5.0 (Server edition)
--------------------------------------------------

-- Main RasChess functions and tables

-- 	RasChess:NewGame(shub,nick1,nick2,chessgame)
--		returns a ChessGameTable

--	RasChess:DelGame(nick)
--		returns 1 if a game was deleted else 'nil'

-- 	RasChess:GetGame(nick)
--		returns a game if one is found else 'nil'

-- Function for the sgame = RasChess:GetGame(nick)

--	sgame:Move(from1,from2,to1,to2,curnick,mode,tofig)
--		return curgame,ret,errl,errc
--		curgame is a copy of sgame
--		Affects sgame only if sgame:doMove(..) ~= "Failed"

--	sgame:doMove(from1,from2,to1,to2,curnick,mode,tofig)
--		returns ret,errl,errc
--		ret either "FAILED","MOVED","CHECK","CHECKMATE" or 'nil',  errl and errc only with "FAILED" but not neccessarily
--		Affects game in any case

--	sgame:SendToPlayers("string")
--		Sends string to Players

--	sgame:SendToPlayerTop("string")
-- 		Sends string to Top Player

--	sgame:sendToPlayerBot("string")
--		Sends string to Bottom Player

--	sgame:SendBoard()
--		Sends the current board to both players

--	sgame:SetFont(nick, font)
--		Set font of nick, returns 1 on success else 'nil'

--	sgame:GetBoard(font)
--		Returns Board with given font

--	sgame:UndoLastMove(nick)
--		Undos Last MOve if nick was last player, If success return 1 else 'nil'

--	sgame:SwitchPlayers()
--		Switches Players

--	sgame:SetGame(curgame)
--		Will assign curgame to sgame




-- Global ChessTable
RasChess = {}

-- Fonts :)

RasChess._Fonts = {}

RasChess._Fonts.Text = 1
RasChess._Fonts.SansSerif = 1

RasChess._Fonts._default = "SansSerif"


-- Table for Players
RasChess._Players = {}

-- Table for Games
RasChess._Games = {}


-- Open a new game (nick1 = opponent),(nick2 = ownnick)
function RasChess:NewGame(shub,nick1,nick2,chessgame)

	-- Get a New Game Index
	local sindex = ""
	local notfound = 1
	while notfound do
		sindex = sindex..string.char(math.random(48,122))
		if not self._Games[sindex] then
			notfound = nil
		end
	end

	-- Check if no specific chessgame should be used
	if (not chessgame) then
		self._Games[sindex] = self:CreateChessGame(nick1,nick2,shub)

		self._Games[sindex].Index = sindex
		
	else
		-- assign SendData To Players func
		chessgame.SendToPlayers = self._GFuncs.SendToPlayers

		-- assign SendBoard to game
		chessgame.SendBoard = self._GFuncs.SendBoard

		chessgame.SendToPlayerTop = self._GFuncs.SendToPlayerTop
		chessgame.SendToPlayerBot = self._GFuncs.SendToPlayerBot

		chessgame.doMove = self._GFuncs.doMove
		chessgame.Move = self._GFuncs.Move
		chessgame.SetFont = self._GFuncs.SetFont
		chessgame.GetBoard = self._GFuncs.GetBoard
		chessgame.UndoLastMove = self._GFuncs.UndoLastMove
		chessgame.SwitchPlayers = self._GFuncs.SwitchPlayers

		chessgame.SetGame = self._GFuncs.SetGame

		chessgame.hub = shub

		chessgame.Index = sindex

		self._Games[sindex] = chessgame
	end

	-- Build references from PlayersTable to the ChessGame
	self._Players[nick1] = {
			index = sindex,
			playswith = nick2,
	}
	self._Players[nick2] = {
			index = sindex,
			playswith = nick1,
	}

	return self._Games[sindex]
	
end


-- Delete a game
function RasChess:DelGame(nick)

	if self._Players[nick] then

		local oppnick = self._Players[nick].playswith
		local gameindex = self._Players[nick].index


		local sgame = self._Games[gameindex]

		-- Delete the GameServer Game
		GameServer:DelGame(sgame.hub,{nick,oppnick})


		-- Delete the ChessGame
		self._Games[gameindex] = nil

		-- Delete Player Entries
		self._Players[nick] = nil
		self._Players[oppnick] = nil

		return 1
	end
end

-- Gets a Playersgame
function RasChess:GetGame(nick)
	if self._Players[nick] then
		local gameindex = self._Players[nick].index
		return self._Games[gameindex]
	end
end

-------------------------------------
--		Create Chess Game
-------------------------------------
-- Figures of the chessgame

RasChess._Figures = {}

RasChess._Figures.White = {}

RasChess._Figures.White.King = { "King","W" }
RasChess._Figures.White.Queen = { "Queen","W" }
RasChess._Figures.White.Bishop = { "Bishop","W" }
RasChess._Figures.White.Horse = { "Horse","W" }
RasChess._Figures.White.Castle = { "Castle","W" }
RasChess._Figures.White.Pawn = { "Pawn","W" }

RasChess._Figures.Black = {}

table.foreach(RasChess._Figures.White, function(i,v)
	RasChess._Figures.Black[i] = { v[1],"B" }
end)


function RasChess:CreateChessGame(nick1,nick2,shub)
	local ChessField = {}

	ChessField.doMove = self._GFuncs.doMove
	ChessField.Move = self._GFuncs.Move

	ChessField.hub = shub

	ChessField.SendToPlayerTop = self._GFuncs.SendToPlayerTop
	ChessField.SendToPlayerBot = self._GFuncs.SendToPlayerBot

	-- assign SendData to game
	ChessField.SendToPlayers = self._GFuncs.SendToPlayers

	-- assign SendBoard to game
	ChessField.SendBoard = self._GFuncs.SendBoard

	-- assign Get Board
	ChessField.GetBoard = self._GFuncs.GetBoard

	-- assign Set Font
	ChessField.SetFont = self._GFuncs.SetFont

	-- assign UndoLastMove
	ChessField.UndoLastMove = self._GFuncs.UndoLastMove

	-- assign SwitchPlayers
	ChessField.SwitchPlayers = self._GFuncs.SwitchPlayers

	-- assign setgame
	ChessField.SetGame = self._GFuncs.SetGame

	-- table for moves
	ChessField.Moves = {}

	-- table for playerfonts
	ChessField.Font = {}
	ChessField.Font[nick1] = self._Fonts._default
	ChessField.Font[nick2] = self._Fonts._default

	ChessField.sides = { Top = nick1, Bottom = nick2 }
	ChessField.players = { nick1,nick2 }
	ChessField.KingMove = {}
	ChessField.KingMove[nick1] = {}
	ChessField.KingMove[nick1].Castle1 = 0
	ChessField.KingMove[nick1].King = 0
	ChessField.KingMove[nick1].Castle2 = 0

	ChessField.KingMove[nick2] = {}
	ChessField.KingMove[nick2].Castle1 = 0
	ChessField.KingMove[nick2].King = 0
	ChessField.KingMove[nick2].Castle2 = 0

	for i = 1,8 do
		for i2 = 1,8 do
			ChessField[i..i2] = { posession = "",ftype = { "","" } }
		end
	end

	-- Set Figs

	local tfigs = self._Figures


	for i = 1,8 do
		ChessField["2"..i].ftype = tfigs.White.Pawn
		ChessField["2"..i].posession = nick1
		ChessField["7"..i].ftype = tfigs.Black.Pawn
		ChessField["7"..i].posession = nick2
	end

	ChessField["11"].ftype = tfigs.White.Castle
	ChessField["11"].posession = nick1
	ChessField["18"].ftype = tfigs.White.Castle
	ChessField["18"].posession = nick1
	ChessField["81"].ftype = tfigs.Black.Castle
	ChessField["81"].posession = nick2
	ChessField["88"].ftype = tfigs.Black.Castle
	ChessField["88"].posession = nick2

	ChessField["12"].ftype = tfigs.White.Horse
	ChessField["12"].posession = nick1
	ChessField["17"].ftype = tfigs.White.Horse
	ChessField["17"].posession = nick1
	ChessField["82"].ftype = tfigs.Black.Horse
	ChessField["82"].posession = nick2
	ChessField["87"].ftype = tfigs.Black.Horse
	ChessField["87"].posession = nick2

	ChessField["13"].ftype = tfigs.White.Bishop
	ChessField["13"].posession = nick1
	ChessField["16"].ftype = tfigs.White.Bishop
	ChessField["16"].posession = nick1
	ChessField["83"].ftype = tfigs.Black.Bishop
	ChessField["83"].posession = nick2
	ChessField["86"].ftype = tfigs.Black.Bishop
	ChessField["86"].posession = nick2

	ChessField["14"].ftype = tfigs.White.King
	ChessField["14"].posession = nick1
	ChessField["15"].ftype = tfigs.White.Queen
	ChessField["15"].posession = nick1
	ChessField["84"].ftype = tfigs.Black.King
	ChessField["84"].posession = nick2
	ChessField["85"].ftype = tfigs.Black.Queen
	ChessField["85"].posession = nick2

	return(ChessField)
end



--------------
-- Funcs for sgame
--------------

RasChess._GFuncs = {}

function RasChess._GFuncs.SendToPlayerTop(self,data,fromnick)
	local nickt = self.sides.Top
	if fromnick then
		local fnick = self.sides.Bottom
		GameServer:SendToNicks(self.hub,{nickt},data,fnick)
	else
		GameServer:SendToNicks(self.hub,{nickt},data)
	end
end

function RasChess._GFuncs.SendToPlayerBot(self,data,fromnick)
	local nickt = self.sides.Bottom
	if fromnick then
		local fnick = self.sides.Top
		GameServer:SendToNicks(self.hub,{nickt},data,fnick)
	else
		GameServer:SendToNicks(self.hub,{nickt},data)
	end
end

-- function SendData To Players
function RasChess._GFuncs.SendToPlayers(self,data)
	self:SendToPlayerTop(data)
	self:SendToPlayerBot(data)
end

-- function SendBoard To Players
function RasChess._GFuncs.SendBoard(self)
	local nickt = self.sides.Top
	local nickb = self.sides.Bottom

	local fontt = self.Font[nickt]
	local fontb = self.Font[nickb]

	if (fontt == fontb) then
		local board = self:GetBoard(fontt)
		self:SendToPlayers(board)
	else
		local boardt = self:GetBoard(fontt)
		self:SendToPlayerTop(boardt)

		local boardb = self:GetBoard(fontb)
		self:SendToPlayerBot(boardb)
	end
end
-- function GetBoard(font)
function RasChess._GFuncs.GetBoard(self,font)
	local sboard = RasChess["_"..font].DrawBoard(self)
	return sboard
end

-- function SetFont
function RasChess._GFuncs.SetFont(self,nick,font)
	if RasChess._Fonts[font] then
		self.Font[nick] = font
		return 1
	end
end

-- function UndoLastMove
function RasChess._GFuncs.UndoLastMove(self,nick)
	if (self.players[2] == nick) then
		if self.Moves[1] then
			for i = 1,table.getn(self.Moves[1]) do
				local directions = self.Moves[1][i].direc
				local from1,from2,to1,to2 = directions[1],directions[2],directions[3],directions[4]
				self[from1..from2] = self.Moves[1][i].from
				self[to1..to2] = self.Moves[1][i].to
			end
			table.remove(self.Moves,1)
		end
		self:SwitchPlayers()
		return 1
	end
end

-- function SwitchPlayers
function RasChess._GFuncs.SwitchPlayers(self)

	local nick1 = self.players[1]
	local nick2 = self.players[2]
	self.players[1] = nick2
	self.players[2] = nick1

end

-- function SetGame
function RasChess._GFuncs.SetGame(self, sgame)
	-- Get GameIndex
	local gameindex = self.Index
	-- Set Index by given Game
	RasChess._Games[gameindex] = sgame
end

-- function Move
function RasChess._GFuncs.Move(self,from1,from2,to1,to2,curnick,mode,tofig)

	-- Create a copy of self
	local curgame = RasChess:tcopy(self)

	local ret,errl,errc = curgame:doMove(from1,from2,to1,to2,curnick,mode,tofig)

	if ret ~= "FAILED" then

		self:SetGame(curgame)
	end

	self = curgame

	return curgame,ret,errl,errc
end

-------------------------------------
--		do a MOve
-------------------------------------
-- mode : 1 = also check for checkmate, 2 = don't check for checkmate
function RasChess._GFuncs.doMove(self,from1,from2,to1,to2,curnick,mode,tofig)

	-- Check if its Players turn
	if (self.players[1] == curnick) then

		-- Check if Fields Fit
		if (self[from1..from2].posession == curnick) and (self[to1..to2].posession ~= curnick) then
	
			-- Check Special Moves of Figures
			if (self[from1..from2].ftype[1] == "King") and (self.KingMove[curnick].King == 0) and (from1 == to1) then

				if ((from2-to2) == 2)
					and (self.KingMove[curnick].Castle1 == 0)
					and (RasChess:doCheckPosRou(self,from1,(from2-1),to2,-1) == "MOVE")
					and (RasChess:doCheckPosRou(self,from1,2,3,1) == "MOVE") then

					-- Move King, Move Castle
					RasChess:moveFigure(self,{{from1,from2,to1,to2},{from1,1,to1,3}})

					local ret,errl,errc = RasChess:doCheckGame(self,from1,from2,to1,to2,curnick,mode)
	
					return ret,errl,errc

				elseif ((to2-from2) == 2)
					and (self.KingMove[curnick].Castle2 == 0)
					and (RasChess:doCheckPosRou(self,from1,(from2+1),to2,1) == "MOVE")
					and (RasChess:doCheckPosRou(self,from1,7,5,-1) == "MOVE") then

					-- Move King
					RasChess:moveFigure(self,{{from1,from2,to1,to2},{from1,8,to1,5}},tofig)

					local ret,errl,errc = RasChess:doCheckGame(self,from1,from2,to1,to2,curnick,mode)

					return ret,errl,errc

	 			end
			end

			-- Check Normal Moves of Figures
			local ret,errl,errc = RasChess._Moves[self[from1..from2].ftype[1]](self,from1,from2,to1,to2)
			if (ret == "MOVE") then

				RasChess:moveFigure(self,{{from1,from2,to1,to2}})

				local ret,errl,errc = RasChess:doCheckGame(self,from1,from2,to1,to2,curnick,mode)

				return ret,errl,errc
			
			elseif (ret == "FAILED") then

				return "FAILED",errl,errc
			end
		end
	end
end


-- Handling Functions
-----------------------
-----------------------

-------------------------------------
--		checkPosess
-------------------------------------
function RasChess:doCheckPosRou(curgame,line,from,to,opt)
	for i = from,to,opt do
		if (curgame[line..i].posession ~= "") then
			return("FAILED")
		end
	end
	return("MOVE")
end

-------------------------------------
--		switch Players
-------------------------------------
function RasChess:switchPlayers(curgame)

	local nick1 = curgame.players[1]
	local nick2 = curgame.players[2]
	curgame.players[1] = nick2
	curgame.players[2] = nick1

end

-------------------------------------
--		move Figure
-------------------------------------
function RasChess:moveFigure(curgame,tomove,tofig)

	local tofig = tofig or "Queen"

	-- Insert Table to Moves
	table.insert(curgame.Moves,1,{})

	for i = 1,table.getn(tomove) do

		-- Get Directions
		local from1,from2,to1,to2 = tomove[i][1],tomove[i][2],tomove[i][3],tomove[i][4]

		-- Insert Moves
		table.insert(curgame.Moves[1],{direc = { from1,from2,to1,to2 },from = curgame[from1..from2],to = curgame[to1..to2]})

		--Set Pawn to Queen if reached other side
		if (curgame[from1..from2].ftype[1] == "Pawn") and ((to1 == 8) or (to1 == 1)) then
			curgame[to1..to2].posession = curgame.players[1]
			if (curgame[from1..from2].ftype[2] == "W") then
				
				-- Check if Fig exists
				if not self._Figures.White[tofig] then
					-- Use Queen as default
					curgame[from1..from2].ftype = self._Figures.White.Queen

				else

					curgame[from1..from2].ftype = self._Figures.White[tofig]

				end

			elseif (curgame[from1..from2].ftype[2] == "B") then

				-- Check if Fig exists
				if not self._Figures.Black[tofig] then
					-- Use Queen as default
					curgame[from1..from2].ftype = self._Figures.Black.Queen

				else

					curgame[from1..from2].ftype = self._Figures.Black[tofig]

				end
			end
		end
		
		curgame[to1..to2] = curgame[from1..from2]
		curgame[from1..from2] = { posession = "",ftype = { "","" } }
	end
	
	return(curgame)
end

-------------------------------------
--		undo Last Move
-------------------------------------
function RasChess:undoLastMove(curgame)
	if curgame.Moves[1] then
		for i = 1,table.getn(curgame.Moves[1]) do
			local directions = curgame.Moves[1][i].direc
			local from1,from2,to1,to2 = directions[1],directions[2],directions[3],directions[4]
			curgame[from1..from2] = curgame.Moves[1][i].from
			curgame[to1..to2] = curgame.Moves[1][i].to
		end
		table.remove(curgame.Moves,1)
		self:switchPlayers(curgame)
	end
end

-------------------------------------
--		do Check for Check Mate
-------------------------------------
function RasChess:doCheckCheckMate(curgame,type)

	local curgame2 = self:tcopy(curgame)

	local otherFigs = {}

	for i = 1,8 do
		for i2 = 1,8 do
			if (curgame2[i..i2].posession == curgame2.players[2]) and (type == "ACT") then
				table.insert(otherFigs, { i,i2 })
			elseif (curgame2[i..i2].posession == curgame2.players[1]) and (type == "PAS") then
				table.insert(otherFigs, { i,i2 })
			end
		end
	end

	self:switchPlayers(curgame2)

	for i = 1,table.getn(otherFigs) do
		local iL1 = otherFigs[i][1]
		local iC1 = otherFigs[i][2]
		for i2 = 1,8 do
			for i3 = 1,8 do

				local ret = curgame2:doMove(iL1,iC1,i2,i3,curgame2.players[1],2)

				if (ret == "CHECK") or (ret == "MOVED") then
					
					return("NOCHECKMATE")

				end
			end
		end
	end

	return("CHECKMATE")

end

-------------------------------------
--		do Check for Check
-------------------------------------
function RasChess:doCheckCheck(curgame)

	local curnick = curgame.players[1]
	local sKings = {}
	sKings[curnick] = {}
	sKings.other = {}
	local sFigs = {}
	sFigs[curnick] = {}
	sFigs.other = {}
	
	for i = 1,8 do
		for i2 = 1,8 do
			if (curgame[i..i2].posession ~= "") then
				if (curgame[i..i2].posession == curnick) then
					sFigs[curnick][i..i2] = 1
					if (curgame[i..i2].ftype[1] == "King") then
						sKings[curnick] = { i,i2 }
					end
				else
					sFigs.other[i..i2] = 1
					if (curgame[i..i2].ftype[1] == "King") then
						sKings.other = { i,i2 }
					end
				end
			end
		end
	end

	-- Check if opponets figs can reach king within next move
	for i,_ in sFigs.other do
		local L1,C1 = tonumber(string.sub(i,1,1)),tonumber(string.sub(i,2,2))
		if (self._Moves[curgame[i].ftype[1]](curgame,L1,C1,sKings[curnick][1],sKings[curnick][2]) == "MOVE") then
			return "FAILED",L1,C1
		end
	end

	-- Check if opponets king can be reached by figs within next move
	for i,_ in sFigs[curnick] do
		local L1,C1 = tonumber(string.sub(i,1,1)),tonumber(string.sub(i,2,2))
		if (self._Moves[curgame[i].ftype[1]](curgame,L1,C1,sKings.other[1],sKings.other[2]) == "MOVE") then	
			return "CHECK",L1,C1
		end
	end

	return("PASSED")
end

-------------------------------------
--		do Global Check of CurGame
-------------------------------------
function RasChess:doCheckGame(curgame,from1,from2,to1,to2,curnick,mode)
	
	local ret,errl,errc = self:doCheckCheck(curgame)

	if (ret == "PASSED") then

		if (mode == 1) then
			--Setting the King or Castel to moved
			if (curgame[from1..from2].ftype[1] == "King") then
				curgame.KingMove[curnick].King = 1
			elseif (curgame[from1..from2].ftype[1] == "Castle") and ((from1==1 or from1==8) and from2 == 1) then
				curgame.KingMove[curnick].Castle1 = 1
			elseif (curgame[from1..from2].ftype[1] == "Castle") and ((from1==1 or from1==8) and from2 == 8) then
				curgame.KingMove[curnick].Castle2 = 1
			end
			if (self:doCheckCheckMate(curgame,"ACT") == "CHECKMATE") then
				return "CHECKMATE"
			end
		end

		self:switchPlayers(curgame)
		
		return "MOVED"

	elseif (ret == "CHECK") then

		if (mode == 1) then
			--Setting the King or Castel to moved
			if (curgame[from1..from2].ftype[1] == "King") then
				curgame.KingMove[curnick].King = 1
			elseif (curgame[from1..from2].ftype[1] == "Castle") and ((from1==1 or from1==8) and from2 == 1) then
				curgame.KingMove[curnick].Castle1 = 1
			elseif (curgame[from1..from2].ftype[1] == "Castle") and ((from1==1 or from1==8) and from2 == 8) then
				curgame.KingMove[curnick].Castle2 = 1
			end

			if (self:doCheckCheckMate(curgame,"ACT") == "CHECKMATE") then
				return "CHECKMATE"
			end
		end
		
		self:switchPlayers(curgame)
		
		return "CHECK",errl,errc

	elseif (ret == "FAILED") then

		return "FAILED",errl,errc

	end
end


-------------------------------------
--		http://lua-users.org/wiki/PitLibTablestuff, modded to only reference objekt hub
-------------------------------------
function RasChess:tcopy(t, lookup_table)
	local copy = {}
	for i,v in t do
		-- the mod
		if i == "hub" then
			copy[i] = v

		elseif type(v) ~= "table" then
			copy[i] = v

		else

			lookup_table = lookup_table or {}

			lookup_table[t] = copy

			if lookup_table[v] then
				copy[i] = lookup_table[v] -- we already copied this table. reuse the copy.
			else
				copy[i] = self:tcopy(v,lookup_table) -- not yet copied. copy it.
			end
		end
	end
	return copy
end
