--// vim:ts=4:sw=4:noet

-- use a namespace because we are in a global environment
formatting = {}

-- load pickle functions
dofile("scripts/libsimplepickle.lua")
formatting.settings_file = "scripts/formatting_settings.txt"
	
-- every new line starts with the following, because of the multiline-URL-fix-hack
formatting.input_start = " "

-- regexes that should be considered noise
formatting.noise = {
	-- NMDC Kick
	"^"..formatting.input_start.."\r\n.* <[^> ]+> [^ ]+ is kicking [^ ]+ because",
	-- DCH++ Banuser/Bannick/Banip
	"^"..formatting.input_start.."\r\n.* <[^> ]+> [^ ]+ banned [^ ]+ by ip [^ ]+ and nick %(",
	"^"..formatting.input_start.."\r\n.* <[^> ]+> [^ ]+ banned [0-9.]+ %(",
	"^"..formatting.input_start.."\r\n.* <[^> ]+> [^ ]+ banned [^ ]+ by nick %(",
	-- YHub Kick
	"^"..formatting.input_start.."\r\n.* <[^> ]+> [^ ]+, IP: [0-9.]+ was kicked",
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*<[^> ]+> [^ ]+ was kicked from the hub",
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*<[^> ]+> [^ ]+ was banned",
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*<[^> ]+> [^ ]+, IP address: [^ ]+ was banned by [^ ]+ because:",
	-- Verihub kick
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*<[^> ]+> Kicked user [^ ]+ IP: [0-9.]+",
	-- More Kicks
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*<[^> ]+> [^ ]+, IP address: [^ ]+ was kicked by [^ ]+ because:",
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*<[^> ]+> [^ ]+, IP address: [^ ]+ was kicked %(total kicks: [^ ]+%) by [^ ]+ because:",
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*<[^> ]+> [^ ]+, IP address: [^ ]+ was banned for [^ ]+ hours by [^ ]+ because:",
	-- ADC test commands
	"^"..formatting.input_start.."\r\n.* <ADC>.*</ADC>$",
	-- p14nd4's join/part noise add
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*** Joins: [^ ]",
	"^"..formatting.input_start.."\r\n[0-9:%[%] ]*** Parts: [^ ]",
}

function dcpp.FormatChatText( hub, text )
	local txtcolor = formatting.color.normal
	local timecolor = formatting.color.time
	local nickcolor = formatting.color.nick
	-- Gandalf start
	local countrycodecolor = formatting.color.countrycode
  local ipcolor = formatting.color.ip
	local mynickcolor = formatting.color.myname
	-- Gandalf end

	if formatting.seconds_in_timestamp then
		local ret,c,rest = string.find( text, "^"..formatting.input_start.."\r\n%[[0-9][0-9]:[0-9][0-9]%](.*)$" )
		if ret then text = formatting.input_start..os.date("\r\n[%H:%M:%S]")..rest end
	end		
	
	if formatting.isNoise( text ) then
		-- escape chars
		text = string.gsub( text, "([{}\\])", "\\%1" )
		text = string.gsub( text, "\r\n", "\n" )
		text = string.gsub( text, "\n", "\\line\n" )
		txtcolor = formatting.color.noise
	elseif dcpp:hasHub( hub ) and dcpp:getHub( hub ):getOwnNick() then
		local hub = dcpp:getHub( hub )
		local mynick = hub:getOwnNick()
		if formatting.testMyLine( text, mynick ) then
			txtcolor = formatting.color.self
			formatting.findChatPartner( hub, text )
			formatting.match_nick_ok = 0
		-- Gandalf start
		elseif formatting_opt.myname > 0 then
			if formatting.testMyNick( text, mynick ) then
				if formatting_opt.myname == 1 then			-- NICK --
					formatting.match_nick = get_match_nick( text, mynick )										
				end
				if formatting_opt.myname > 1 then				-- MSG, ROW --
					txtcolor = formatting.color.myname
				end
				if formatting_opt.myname == 3 then			-- ROW --
					timecolor = formatting.color.myname
					countrycodecolor = formatting.color.myname
					ipcolor = formatting.color.myname
					nickcolor = formatting.color.myname
				end
				if formatting.flash_on_myname then
					hub:attention()
				end
			end
			-- Gandalf end
		elseif formatting.testPartnerLine( text ) then
			txtcolor = formatting.color.partner
		end
		-- escape chars
		text = string.gsub( text, "([{}\\])", "\\%1" )
		text = string.gsub( text, "\r\n", "\n" )
		text = string.gsub( text, "\n", "\\line\n" )
		-- colorize OPs
		local ret,c,pre,nick,post = string.find( text,
				"^("..formatting.input_start.."\\line\n[0-9A-Z:.%[%] ]*<)([^> ]+)(>.*)$" )
		if ret then
			-- Gandalf start
			if formatting_opt.bold == 1 then
				text = string.sub( pre, 1, -2 ).."\\cf3 <\\b "..nick.."\\b0 >\\cf1 "..string.sub( post, 2 )
			else
				text = string.sub( pre, 1, -2 ).."\\cf3 <"..nick..">\\cf1 "..string.sub( post, 2 )
			end
			-- Gandalf end

			if hub:isOp( nick ) then
				nickcolor = formatting.color.op
			end
		end
	else
		-- escape chars
		text = string.gsub( text, "([{}\\])", "\\%1" )
		text = string.gsub( text, "\r\n", "\n" )
		text = string.gsub( text, "\n", "\\line\n" )
	end
	
	-- colorize timestamp
	text = string.gsub( text, "^("..formatting.input_start.."\\line\n)([%[0-9A-Z:. %]]+)", "%1\\cf4 %2\\cf1 " )
	-- Gandalf start
	-- colorize country code
	text = string.gsub( text, "^("..formatting.input_start.."\\line\n.-)(%[ %u%u %]+)", "%1\\cf5 %2\\cf1 " )
	-- colorize ip
	text = string.gsub( text, "^("..formatting.input_start.."\\line\n.-)(%[ %d+.%d+.%d+.%d+ %]+)", "%1\\cf6 %2\\cf1 " )
	-- colorize mynick
	if formatting_opt.myname == 1 and formatting.match_nick_ok == 1 then	-- NICK --
		text = string.gsub( text, "^("..formatting.input_start.."\\line\n.-)("..formatting.match_nick..")", "%1\\cf7 %2\\cf1 " )
	end
	--  Gandalf end
	-- em
	text = string.gsub( text, "(%s)(_%S+_)$", "%1\\ul %2\\ul0 " )
	text = string.gsub( text, "(%s)(_%S+_)(%s)", "%1\\ul %2\\ul0 %3" )

	text = string.gsub( text, "(%s)(*%S+*)$", "%1\\b %2\\b0 " )
	text = string.gsub( text, "(%s)(*%S+*)(%s)", "%1\\b %2\\b0 %3" )

	text = string.gsub( text, "(%s)(/[^/\r\n\t]+/)$", "%1\\i %2\\i0 " )
	text = string.gsub( text, "(%s)(/[^/\n\n\t]+/)(%s)", "%1\\i %2\\i0 %3" )

	-- colorize links
	text = string.gsub( text, "([ <\t\r\n])(dchub://[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(http://[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(https://[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(www\.[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(ftp://[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(magnet:%?[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(irc://[^ \t\\>]+)","%1\\ul\\cf2 %2\\cf1\\ul0 " )
	
	-- Gandalf start
	--return "{\\rtf1\\ansi\\deff0\\plain0\n"..
	--		"{\\colortbl ;"..txtcolor..formatting.color.link..nickcolor..timecolor.."}\n"..
	--		"\\cf1 "..text.."}\n"

	return "{\\rtf1\\ansi\\deff0\\plain0\n"..
			"{\\colortbl ;"..txtcolor..formatting.color.link..nickcolor..timecolor..
			countrycodecolor..ipcolor..mynickcolor.."}\n"..
			"\\cf1 "..text.."}\n"
	-- Gandalf end
end

-- time data
formatting.time = {}
formatting.time.partners = {}

-- color conversion types
formatting.color_conversion = {}
formatting.color_conversion.black	= {0, 0, 0}
formatting.color_conversion.white	= {255, 255, 255}
formatting.color_conversion.gray	= {127, 127, 127}
formatting.color_conversion.red		= {255, 0, 0}
formatting.color_conversion.green	= {0, 255, 0}
formatting.color_conversion.blue	= {0, 0, 255}
formatting.color_conversion.yellow	= {255, 255, 0}
formatting.color_conversion.magenta	= {255, 0, 255}
formatting.color_conversion.cyan	= {0, 255, 255}
-- dark variants
formatting.color_conversion.darkgray	= {63, 63, 63}
formatting.color_conversion.darkred		= {127, 0, 0}
formatting.color_conversion.darkgreen	= {0, 127, 0}
formatting.color_conversion.darkblue	= {0, 0, 127}
formatting.color_conversion.darkyellow	= {127, 127, 0}
formatting.color_conversion.darkmagenta	= {127, 0, 127}
formatting.color_conversion.darkcyan	= {0, 127, 127}
-- aliases
formatting.color_conversion.brown		= formatting.color_conversion.darkyellow
formatting.color_conversion.grey		= formatting.color_conversion.gray
formatting.color_conversion.darkgrey	= formatting.color_conversion.darkgray
formatting.color_conversion.lightgrey  = formatting.color_conversion.lightgray

-- convert colors..
function formatting.setColors()
	formatting.color = {}
	for k,v in formatting_opt.color do
		if formatting.color_conversion[v] then
			local c = formatting.color_conversion[v]
			formatting.color[k] = "\\red"..c[1].."\\green"..c[2].."\\blue"..c[3]..";"
		else
			formatting.color[k] = v
		end
		-- optionally invert color
		if formatting.background_is_dark then
			local ret,c,r,g,b = string.find( formatting.color[k], "^\\red(%d+)\\green(%d+)\\blue(%d+);$" )
			if ret then
				formatting.color[k] = "\\red"..(255-tonumber(r))..
										"\\green"..(255-tonumber(g))..
										"\\blue"..(255-tonumber(b))..";"
			end
		end
	end
end

-- Gandalf start
-- convert pmcolors..
function formatting.setPmColors()
	formatting.pmcolor = {}
	for k,v in formatting_opt.pmcolor do
		if formatting.color_conversion[v] then
			local c = formatting.color_conversion[v]
			formatting.pmcolor[k] = "\\red"..c[1].."\\green"..c[2].."\\blue"..c[3]..";"
		else
			formatting.pmcolor[k] = v
		end
		-- optionally invert color
		if formatting.background_is_dark then
			local ret,c,r,g,b = string.find( formatting.pmcolor[k], "^\\red(%d+)\\green(%d+)\\blue(%d+);$" )
			if ret then
				formatting.pmcolor[k] = "\\red"..(255-tonumber(r))..
										"\\green"..(255-tonumber(g))..
										"\\blue"..(255-tonumber(b))..";"
			end
		end
	end
end
-- Gandalf end

-- set "also match nicks"
function formatting.setAlsoMatchNicks()
	formatting.also_match_nicks = {}
	for k,v in formatting_opt.also_match_nicks do
		formatting.also_match_nicks[k] = string.gsub( string.lower( v ), "([^%l%d])", "%%%1" )
	end
end

-- set other options
function formatting.setOtherOptions()
	local function test( b )
		b = string.lower( tostring( b ) )
		if b == "nil" or b == "0" or b == "off" or b == "no" then return nil
		else return 1
		end
	end
	formatting.flash_on_myname = test( formatting_opt.flash_on_myname )
	formatting.background_is_dark = test( formatting_opt.background_is_dark )
	formatting.seconds_in_timestamp = test( formatting_opt.seconds_in_timestamp )
	-- Gandalf start
  formatting.partner_time = formatting_opt.partner_time
	-- Gandalf end
end

function formatting.setOptions()
	formatting.setDefaultOptions()
	formatting.setOtherOptions()
	formatting.setColors()
	-- Gandalf start
  formatting.setPmColors()
  -- Gandalf end
	formatting.setAlsoMatchNicks()
end

function formatting.setDefaultOptions()
	if not formatting_opt then formatting_opt = {} end
	if not formatting_opt.background_is_dark then
		local bgcolor = DC():GetSetting( "BackgroundColor" ) -- BACKGROUND_COLOR (IntSetting)
		local R = math.mod( bgcolor, 256 )
		-- Gandalf start
		--local G = math.mod( bgcolor / 256, 256 )
		--local B = math.mod( bgcolor / 256 / 256, 256 )
		local G = math.mod( math.floor(bgcolor / 256), 256 )
		local B = math.mod( math.floor(bgcolor / 256 / 256), 256 )
		-- Gandalf end
		if R + G + B < 384 then
			formatting_opt.background_is_dark = 1
		else
			formatting_opt.background_is_dark = 0
		end
	end
	if not formatting_opt.seconds_in_timestamp then formatting_opt.seconds_in_timestamp = 1 end
	if not formatting_opt.flash_on_myname then formatting_opt.flash_on_myname = 0 end
	if not formatting_opt.color then formatting_opt.color = {} end
	if not formatting_opt.color.normal then formatting_opt.color.normal = "black" end
	if not formatting_opt.color.self then formatting_opt.color.self = "red" end
	if not formatting_opt.color.myname then formatting_opt.color.myname = "darkgreen" end
	if not formatting_opt.color.time then formatting_opt.color.time = "darkgray" end
	-- Gandalf start
  if not formatting_opt.color.countrycode then formatting_opt.color.countrycode = "darkgray" end
  if not formatting_opt.color.ip then formatting_opt.color.ip = "darkgray" end
  -- Gandalf end
	if not formatting_opt.color.link then formatting_opt.color.link = "blue" end
	if not formatting_opt.color.op then formatting_opt.color.op = "blue" end
	if not formatting_opt.color.nick then formatting_opt.color.nick = "darkred" end
	if not formatting_opt.color.partner then formatting_opt.color.partner = "darkmagenta" end
	if not formatting_opt.color.noise then formatting_opt.color.noise = "darkgray" end
	if not formatting_opt.pmcolor then formatting_opt.pmcolor = {} end
	if not formatting_opt.pmcolor.normal then formatting_opt.pmcolor.normal = "black" end
	if not formatting_opt.pmcolor.self then formatting_opt.pmcolor.self = "red" end
	if not formatting_opt.pmcolor.myname then formatting_opt.pmcolor.myname = "darkgreen" end
	if not formatting_opt.pmcolor.time then formatting_opt.pmcolor.time = "darkgray" end
	-- Gandalf start
  if not formatting_opt.pmcolor.countrycode then formatting_opt.pmcolor.countrycode = "darkgray" end
  if not formatting_opt.pmcolor.ip then formatting_opt.pmcolor.ip = "darkgray" end
  -- Gandalf end
	if not formatting_opt.pmcolor.link then formatting_opt.pmcolor.link = "blue" end
	if not formatting_opt.pmcolor.op then formatting_opt.pmcolor.op = "blue" end
	if not formatting_opt.pmcolor.nick then formatting_opt.pmcolor.nick = "darkred" end
	if not formatting_opt.pmcolor.partner then formatting_opt.pmcolor.partner = "darkmagenta" end
	if not formatting_opt.pmcolor.noise then formatting_opt.pmcolor.noise = "darkgray" end
	if not formatting_opt.also_match_nicks then formatting_opt.also_match_nicks = {} end
	-- Gandalf start
	if not formatting_opt.bold then formatting_opt.bold = 1 end
	if not formatting_opt.myname then formatting_opt.myname = 1 end
  if not formatting_opt.partner_time then formatting_opt.partner_time = 10 end
  -- Gandalf end
end

function formatting.save()
	formatting.setOptions()
	pickle.store( formatting.settings_file, {formatting_opt = formatting_opt} )
end

function formatting.load()
	local o = io.open( formatting.settings_file, "r" )
	if o then
		dofile( formatting.settings_file )
		o:close()
	end
	formatting.setOptions()
end

--/////////////////////////////////--
--// chatline matching functions //--
--/////////////////////////////////--

function formatting.testMyLine( text, nick )
	if string.find( text, "<"..nick.."> ", 1, 1 ) then
		return 1
	end
end
function formatting.testPartnerLine( text )
	local now = os.time()
	local ret = nil
	for k,v in formatting.time.partners do
		if string.find( text, "<"..k.."> ", 1, 1 ) then
			ret = 1
			formatting.time.partners[k] = now -- re-enforce user speaking
		else
			-- Gandalf
			--if now - v > 60*10 then -- expire after 10 mins
			if now - v > 60*formatting.partner_time then -- expire after mins
			-- Gandalf end
				formatting.time.partners[k] = nil
			end
		end
	end
	return ret
end

function formatting.testMyNick( text, nick )
	text = string.lower( text )
	formatting.match_nick_ok = 0	-- Gandalf start / end

	-- test also_match_nicks first
	for k,v in formatting.also_match_nicks do
		if string.find( text, "[^%l]"..v.."[^%l]" ) or string.find( text, "[^%l]"..v.."$" ) then
			formatting.match_nick_ok = 1	-- Gandalf start / end
			return 1
		end
	end
	
	-- ok.. not found.. try our hub specific nick
	nick = string.lower( nick )
	local enick = nil
	local enick_notag = string.gsub( nick, "%[.*%]", "" )
	if enick_notag ~= "" then
		enick = string.gsub( enick_notag, "([^a-z0-9])", "%%%1" )
	else
		enick = string.gsub( nick, "([^a-z0-9])", "%%%1" )
	end

	if string.find( text, "[^a-z]"..enick.."[^a-z]" ) or string.find( text, "[^a-z]"..enick.."$" ) then
		formatting.match_nick_ok = 1	-- Gandalf start / end
		return 1
	end
end

-- Gandalf start
function get_match_nick( text, nick )
	local ltext = string.lower( text )
	local retnick = ""
	
	-- test also_match_nicks first
	for k,v in formatting.also_match_nicks do
		start,stop = string.find( ltext, v )	
		if start then
			retnick = string.sub( text, start, stop )
			return retnick
		end
	end

	-- ok.. not found.. try our hub specific nick
	local nick = string.lower( nick )
	local nick_notag = string.gsub( nick, "%[.*%]", "" )
	--nick = string.gsub( nick , "%[(.*)%]", "%%[%1%%]" )
	nick = string.gsub( nick, "([%[%]%(%)])", "%%%1" )

	start,stop = string.find( ltext, nick )
	if not start then
		start,stop = string.find( ltext, nick_notag )
	end
	if start then 
		retnick = string.sub( text, start, stop )
		--retnick = string.gsub( retnick , "%[(.*)%]", "%%[%1%%]" )
		retnick = string.gsub( retnick ,  "([%[%]%(%)])", "%%%1" )
	end

	return retnick
end
-- Gandalf end

function formatting.findChatPartner( hub, text )
	local ret,c,nick = string.find( text, "> ([%[%]%a%d_-]+)" )
	if not ret then return end
	local notag = nil
	if not string.find( nick, "[%[%]]" ) then notag = 1 end
	-- iterate through hub nicks to find someone,
	-- if we didn't check user existence, we'd store a heck of a lot of words
	local list = hub:findUsers( nick, notag )
	local now = os.time()
	for k,v in list do
		formatting.time.partners[v:getNick()] = now
	end
end
function formatting.isNoise( text )
	for k,v in formatting.noise do
		if string.find( text, v ) then
			return 1
		end
	end
end


--//////////////////////////////--


function formatting.tokenize( str )
	local ret = {}
	string.gsub( str, "(%w+)", function( s ) table.insert( ret, s ) end )
	return ret
end

dcpp:setListener( "ownChatOut", "formatting",
	function( hub, text )
		if string.sub( text, 1, 1 ) ~= "/" then return end
		parms = formatting.tokenize( string.sub( string.lower( text ), 2 ) )
		if parms[1] == "help" then
		-- Gandalf start
		--	hub:injectChat( "*** (formatting.lua) /color [what] [color], /flash [on/off], "..
		--					"/highlight [add/del] [nick], /invert [on/off], /seconds [on/off]" )
			hub:injectChat( "*** (formatting.lua) /color [what] [color], /pmcolor [what] [color], /flash [on/off], "..
							"/highlight [add/del] [nick], /invert [on/off], /myname [nick/msg/row/off], /mins [minutes], /bold [on/off], /partners" )
      -- Gandalf end	
			return 1
		elseif parms[1] == "seconds" then
			if parms[2] == "on" then
				formatting_opt.seconds_in_timestamp = 1
				formatting.save()
				hub:injectChat( "*** Seconds in timestamp enabled" )
			elseif parms[2] == "off" then
				formatting_opt.seconds_in_timestamp = 0
				formatting.save()
				hub:injectChat( "*** Seconds in timestamp disabled" )
			else
				hub:injectChat( "*** /seconds (parameters: 'on' or 'off'): "..
								"When enabled, BCDC++ displays seconds in the timestamp in main chat." )
			end
			return 1
		-- Gandald start
		elseif parms[1] == "myname" then
			if parms[2] == "nick" then
				formatting_opt.myname = 1
				formatting.save()
				hub:injectChat( "*** Colorize only own nick is enabled" )
			elseif parms[2] == "msg" then
				formatting_opt.myname = 2
				formatting.save()
				hub:injectChat( "*** Colorize only msg is enabled" )
			elseif parms[2] == "row" then
				formatting_opt.myname = 3
				formatting.save()
				hub:injectChat( "*** Colorize hole row is enabled" )
			elseif parms[2] == "off" then
				formatting_opt.myname = 0
				formatting.save()
				hub:injectChat( "*** Colorize myname is disabled" )
			else
				hub:injectChat( "*** /myname (parameters: 'nick', 'msg', 'row' or 'off'): \r\n"..
								"nick\t= only your own nick is colorized whit 'myname' color when your nick is in the message.\r\n"..
								"msg\t= only the message is colorized whit 'myname' color when your nick is in the message.\r\n"..
								"row\t= the hole row is colorized whit 'myname' color when your nick is in the message.\r\n"..			
								"off\t= the 'myname' color is disabled.")
			end
			return 1
    elseif parms[1] == "mins" then
      if not parms[2] then
        hub:injectChat( "*** /mins (parameters: 'mins'): "..
								"This is the time that the user is colored as partner. For the moment the time is set to: "..formatting.partner_time.." minutes." )
      else
			  if string.find(parms[2],"[0-9]+")then
				  formatting_opt.partner_time = parms[2]
				  formatting.save()
				  hub:injectChat( "*** Partner time is now set to "..parms[2].." minutes." )
			  else
				  hub:injectChat( "*** /mins (parameters: 'mins'): "..
								"You have to give a numric value for the minutes." )
			  end
      end
			return 1
		-- Gandalf end
		elseif parms[1] == "flash" then
			if parms[2] == "on" then
				formatting_opt.flash_on_myname = 1
				formatting.save()
				hub:injectChat( "*** Flash on your nick enabled" )
			elseif parms[2] == "off" then
				formatting_opt.flash_on_myname = 0
				formatting.save()
				hub:injectChat( "*** Flash on your nick disabled" )
			else
				hub:injectChat( "*** /flash (parameters: 'on' or 'off'): "..
								"When enabled, the BCDC++ window flashes when someone calls you in main chat." )
			end
			return 1
		elseif parms[1] == "invert" then
			if parms[2] == "on" then
				formatting_opt.background_is_dark = 1
				formatting.save()
				hub:injectChat( "*** Invert colors enabled")
			elseif parms[2] == "off" then
				formatting_opt.background_is_dark = 0
				formatting.save()
				hub:injectChat( "*** Invert colors disabled" )
			else
				hub:injectChat( "*** /invert (parameters: 'on' or 'off'): "..
								"When enabled, the main chat colors are inverted, useful when there's "..
								"low contrast between the chat colors and background (e.g. if you have a "..
								"black background). "..
								"As an alternate option, you can set the colors manually as you like by "..
								"using /color, see /help." )
			end
			return 1
		elseif parms[1] == "color" then
			if parms[2] == "show" then
				-- show current settings
				local msg = "*** The current color set is the following:"
				for k,v in formatting_opt.color do
					msg = msg.. "\r\n*"..k..":* "..v
				end
				hub:injectChat( msg )
			elseif not parms[2] or not parms[3] then
				local msg = "*** /color (parameters: <what> and <color>): "..
							"Use this to change the colors used in the chat.\r\n"..
							"Available color-identifiers to set (<what>):\r\n"..
							"  normal = the normal text color\r\n"..
							"  self = lines you speak\r\n"..
							"  myname = lines where your name is used (someone calling you)\r\n"..
							"  time = the timestamp\r\n"..
							"  link = URLs (clickable links)\r\n"..
							"  op = the nickname of an operator (the one between angle brackets)\r\n"..
							"  nick = the nickname of normal people\r\n"..
							"  partner = lines someone with whom you spoke recently\r\n"..
							"  noise = kick messages\r\n"..
							"Available colors (<color>):\r\n  "
				for k,v in formatting.color_conversion do
					msg = msg..k..", "
				end
				msg = string.sub( msg, 1, -3 )
				msg = msg.."\r\nUse \"/color show\" to get the current color set displayed."
				hub:injectChat( msg )
			else
				if formatting_opt.color[parms[2]] then
					if formatting.color_conversion[parms[3]] then
						formatting_opt.color[parms[2]] = parms[3];
						formatting.save()
						hub:injectChat( "*** /color -- set "..parms[2].." to "..parms[3] )	
					else
						hub:injectChat( "*** /color -- invalid <color>" )
					end
				else
					hub:injectChat( "*** /color -- invalid <what>" )
				end
			end
			return 1
		-- Gandalf start
		elseif parms[1] == "pmcolor" then
			if parms[2] == "show" then
				-- show current settings
				local msg = "*** The current pmcolor set is the following:"
				for k,v in formatting_opt.pmcolor do
					msg = msg.. "\r\n*"..k..":* "..v
				end
				hub:injectChat( msg )
			elseif not parms[2] or not parms[3] then
				local msg = "*** /pmcolor (parameters: <what> and <color>): "..
							"Use this to change the colors used in the chat.\r\n"..
							"Available color-identifiers to set (<what>):\r\n"..
							"  normal\t\t= the normal text color\r\n"..
							"  self\t\t= lines you speak\r\n"..
							"  myname\t\t= lines where your name is used (someone calling you)\r\n"..
							"  time\t\t= the timestamp\r\n"..
							"  link\t\t= URLs (clickable links)\r\n"..
							"  op\t\t= the nickname of an operator (the one between angle brackets)\r\n"..
							"  nick\t\t= the nickname of normal people\r\n"..
							"  partner\t\t= lines someone with whom you spoke recently\r\n"..
							"  noise\t\t= kick messages\r\n"..
              "  countrycode\t= the country code if you use it\r\n"..
              "  ip\t\t= the ip number if you use it\r\n"..
							"Available pmcolors (<color>):\r\n  "
				for k,v in formatting.color_conversion do
					msg = msg..k..", "
				end
				msg = string.sub( msg, 1, -3 )
				msg = msg.."\r\nUse \"/pmcolor show\" to get the current color set displayed."
				hub:injectChat( msg )
			else
				if formatting_opt.pmcolor[parms[2]] then
					if formatting.color_conversion[parms[3]] then
						formatting_opt.pmcolor[parms[2]] = parms[3];
						formatting.save()
						hub:injectChat( "*** /pmcolor -- set "..parms[2].." to "..parms[3] )	
					else
						hub:injectChat( "*** /pmcolor -- invalid <color>" )
					end
				else
					hub:injectChat( "*** /pmcolor -- invalid <what>" )
				end
			end
			return 1
		-- Gandalf end
		elseif parms[1] == "highlight" then
			if not parms[2] or not parms[3] then
				local msg = "*** /highlight (parameters: 'add' or 'del' and <word>): "..
							"Adds or removes words to match for highlighting besides your own name. "..
							"E.g. use \"/highlight add operator\" to let BCDC++ highlight lines that "..
							"include the word \"operator\".\r\n"..
							"Currently matched words:\r\n  "
				for k,v in formatting_opt.also_match_nicks do
					msg = msg..v..", "
				end
				hub:injectChat( string.sub( msg, 1, -3 ) )
			elseif parms[2] == "add" then
				for k,v in formatting_opt.also_match_nicks do
					if string.lower( v ) == parms[3] then
						hub:injectChat( "*** /highlight -- "..parms[3].." exists already" )
						return 1
					end
				end
				table.insert( formatting_opt.also_match_nicks, parms[3] )
				formatting.save()
				hub:injectChat( "*** Highlight, added: "..parms[3] )
			elseif parms[2] == "del" then
				local rm = nil
				for k,v in formatting_opt.also_match_nicks do
					if string.lower( v ) == parms[3] then
						table.remove( formatting_opt.also_match_nicks, k )
						rm = 1
					end
				end
				if rm then
					formatting.save()
					hub:injectChat( "*** Highlight, removed: "..parms[3] )
				else
					hub:injectChat( "*** /highlight -- "..parms[3].." didn't exist" )
				end
			else
				hub:injectChat( "*** /highlight -- "..parms[2].." is not a switch" )
			end	
			return 1
		-- Gandalf start
		elseif parms[1] == "partners" then
			local now = os.time()
			local timeleft
			local timestr
      local msg = "\r\n*** Partner list for the moment....\r\n"
      for k,v in formatting.time.partners do
				timeleft = (formatting.partner_time*60)-os.difftime(now, v)
				timestr = math.floor(timeleft/60)..":"..math.mod(timeleft,60)
        msg = msg..k.." -> "..timestr.." mins left\r\n"
      end
      msg = msg.."*** List end"
      hub:injectChat( msg )
      return 1
		elseif parms[1] == "bold" then
			if parms[2] == "on" then
				formatting_opt.bold = 1
				formatting.save()
				hub:injectChat( "*** Bold nicks enabled" )
			elseif parms[2] == "off" then
				formatting_opt.bold = 0
				formatting.save()
				hub:injectChat( "*** Bold nicks disabled" )
			else
				hub:injectChat( "*** /bold (parameters: 'on' or 'off): " ..
								"When enabled, the nicks are written with bold font to the main chat." )
			end
			return 1
		-- Gandalf end
		end
	end
)

function dcpp.FormatPmText( hub, text )
	local txtpmcolor = formatting.pmcolor.normal
	local timepmcolor = formatting.pmcolor.time
	local nickpmcolor = formatting.pmcolor.nick
	-- Gandalf start
	local countrycodepmcolor = formatting.pmcolor.countrycode
  local ippmcolor = formatting.pmcolor.ip
	local mynickpmcolor = formatting.pmcolor.myname
	-- Gandalf end

	if formatting.seconds_in_timestamp then
		local ret,c,rest = string.find( text, "^"..formatting.input_start.."\r\n%[[0-9][0-9]:[0-9][0-9]%](.*)$" )
		if ret then text = formatting.input_start..os.date("\r\n[%H:%M:%S]")..rest end
	end		
	
	if formatting.isNoise( text ) then
		-- escape chars
		text = string.gsub( text, "([{}\\])", "\\%1" )
		text = string.gsub( text, "\r\n", "\n" )
		text = string.gsub( text, "\n", "\\line\n" )
		txtpmcolor = formatting.pmcolor.noise
	elseif dcpp:hasHub( hub ) and dcpp:getHub( hub ):getOwnNick() then
		local hub = dcpp:getHub( hub )
		local mynick = hub:getOwnNick()
		if formatting.testMyLine( text, mynick ) then
			txtpmcolor = formatting.pmcolor.self
			formatting.findChatPartner( hub, text )
			formatting.match_nick_ok = 0
		-- Gandalf start
		elseif formatting_opt.myname > 0 then
			if formatting.testMyNick( text, mynick ) then
				if formatting_opt.myname == 1 then			-- NICK --
					formatting.match_nick = get_match_nick( text, mynick )										
				end
				if formatting_opt.myname > 1 then				-- MSG, ROW --
					txtpmcolor = formatting.pmcolor.myname
				end
				if formatting_opt.myname == 3 then			-- ROW --
					timepmcolor = formatting.pmcolor.myname
					countrycodepmcolor = formatting.pmcolor.myname
					ippmcolor = formatting.pmcolor.myname
					nickpmcolor = formatting.pmcolor.myname
				end
				if formatting.flash_on_myname then
					hub:attention()
				end
			end
			-- Gandalf end
		elseif formatting.testPartnerLine( text ) then
			txtpmcolor = formatting.color.partner
		end
		-- escape chars
		text = string.gsub( text, "([{}\\])", "\\%1" )
		text = string.gsub( text, "\r\n", "\n" )
		text = string.gsub( text, "\n", "\\line\n" )
		-- colorize OPs
		local ret,c,pre,nick,post = string.find( text,
				"^("..formatting.input_start.."\\line\n[0-9A-Z:.%[%] ]*<)([^> ]+)(>.*)$" )
		if ret then
			-- Gandalf start
			if formatting_opt.bold == 1 then
				text = string.sub( pre, 1, -2 ).."\\cf3 <\\b "..nick.."\\b0 >\\cf1 "..string.sub( post, 2 )
			else
				text = string.sub( pre, 1, -2 ).."\\cf3 <"..nick..">\\cf1 "..string.sub( post, 2 )
			end
			-- Gandalf end
			if hub:isOp( nick ) then
				nickpmcolor = formatting.pmcolor.op
			end
		end
	else
		-- escape chars
		text = string.gsub( text, "([{}\\])", "\\%1" )
		text = string.gsub( text, "\r\n", "\n" )
		text = string.gsub( text, "\n", "\\line\n" )
	end
	
	-- colorize timestamp
	text = string.gsub( text, "^("..formatting.input_start.."\\line\n)([%[0-9A-Z:. %]]+)", "%1\\cf4 %2\\cf1 " )
	-- Gandalf start
	-- colorize country code
  text = string.gsub( text, "^("..formatting.input_start.."\\line\n.-)(%[ %u%u %]+)", "%1\\cf5 %2\\cf1 " )
  -- colorize ip
  text = string.gsub( text, "^("..formatting.input_start.."\\line\n.-)(%[ %d+.%d+.%d+.%d+ %]+)", "%1\\cf6 %2\\cf1 " )
	-- colorize mynick
	if formatting_opt.myname == 1 and formatting.match_nick_ok == 1 then	-- NICK --
		text = string.gsub( text, "^("..formatting.input_start.."\\line\n.-)("..formatting.match_nick..")", "%1\\cf7 %2\\cf1 " )
	end
  --  Gandalf end
	-- em
	text = string.gsub( text, "(%s)(_%S+_)$", "%1\\ul %2\\ul0 " )
	text = string.gsub( text, "(%s)(_%S+_)(%s)", "%1\\ul %2\\ul0 %3" )

	text = string.gsub( text, "(%s)(*%S+*)$", "%1\\b %2\\b0 " )
	text = string.gsub( text, "(%s)(*%S+*)(%s)", "%1\\b %2\\b0 %3" )

	text = string.gsub( text, "(%s)(/[^/\r\n\t]+/)$", "%1\\i %2\\i0 " )
	text = string.gsub( text, "(%s)(/[^/\n\n\t]+/)(%s)", "%1\\i %2\\i0 %3" )

	-- colorize links
	text = string.gsub( text, "([ <\t\r\n])(dchub://[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(http://[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(https://[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(www\.[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(ftp://[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(magnet:%?[^ \t\\>]+)", "%1\\ul\\cf2 %2\\cf1\\ul0 " )
	text = string.gsub( text, "([ <\t\r\n])(irc://[^ \t\\>]+)","%1\\ul\\cf2 %2\\cf1\\ul0 " )

	-- Gandalf start
	--return "{\\rtf1\\ansi\\deff0\\plain0\n"..
	--		"{\\colortbl ;"..txtpmcolor..formatting.pmcolor.link..nickpmcolor..timepmcolor.."}\n"..
	--		"\\cf1 "..text.."}\n"

	return "{\\rtf1\\ansi\\deff0\\plain0\n"..
			"{\\colortbl ;"..txtpmcolor..formatting.pmcolor.link..nickpmcolor..timepmcolor..
      countrycodepmcolor..ippmcolor..mynickpmcolor.."}\n"..
			"\\cf1 "..text.."}\n"
  -- Gandalf end
end

-- load saved settings
formatting.load()
formatting.save() -- recreate settings file

DC():PrintDebug( "  ** Loaded formatting.lua **" )
